port_symmetries_1x1 = {
    "o2@0,o1@0": ["o1@0,o2@0"],
}


port_symmetries_crossing = {
    "o2@0,o1@0": ["o1@0,o2@0", "o3@0,o4@0", "o4@0,o3@0"],
    "o3@0,o1@0": ["o1@0,o3@0", "o2@0,o4@0", "o4@0,o2@0"],
    "o4@0,o1@0": ["o1@0,o4@0", "o2@0,o3@0", "o3@0,o2@0"],
}


if __name__ == "__main__":
    import numpy as np

    port_symmetries = port_symmetries_crossing

    sp = dict(wavelengths=np.linspace(1.5, 1.6, 3))
    sp["o1@0,o1@0"] = 2 * np.linspace(1.5, 1.6, 3)

    for key, symmetries in port_symmetries.items():
        for sym in symmetries:
            if key in sp:
                sp[sym] = sp[key]



"""Classes and utils for working with KLayout technology files (.lyp, .lyt).

This module enables conversion between gdsfactory settings and KLayout technology.
"""

import pathlib
import xml.etree.ElementTree as ET
from typing import Dict, List, Optional, Tuple

from pydantic import BaseModel, Field

from gdsfactory.config import PATH
from gdsfactory.technology import LayerStack, LayerViews
from gdsfactory.typings import PathType

Layer = Tuple[int, int]
ConductorViaConductorName = Tuple[str, str, str]

prefix_d25 = """<?xml version="1.0" encoding="utf-8"?>
<klayout-macro>
 <description/>
 <version/>
 <category>d25</category>
 <prolog/>
 <epilog/>
 <doc/>
 <autorun>false</autorun>
 <autorun-early>false</autorun-early>
 <priority>0</priority>
 <shortcut/>
 <show-in-menu>true</show-in-menu>
 <group-name>d25_scripts</group-name>
 <menu-path>tools_menu.d25.end</menu-path>
 <interpreter>dsl</interpreter>
 <dsl-interpreter-name>d25-dsl-xml</dsl-interpreter-name>
 <text>

"""
suffix_d25 = """
</text>
</klayout-macro>
"""


class KLayoutTechnology(BaseModel):
    """A container for working with KLayout technologies (requires KLayout Python package).

    Useful to import/export Layer Properties (.lyp) and Technology (.lyt) files.

    Properties:
        name: technology name.
        layer_map: Maps names to GDS layer numbers.
        layer_views: Defines all the layer display properties needed for a .lyp file from LayerView objects.
        technology: KLayout Technology object from the KLayout API. Set name, dbu, etc.
        connectivity: List of layer names connectivity for netlist tracing.
    """

    # TODO: Add import method
    # TODO: Also interop with xs scripts?
    import klayout.db as db

    name: str
    layer_map: Dict[str, Layer]
    layer_views: Optional[LayerViews] = None
    layer_stack: Optional[LayerStack] = None
    technology: db.Technology = Field(default_factory=db.Technology)
    connectivity: Optional[List[ConductorViaConductorName]] = None

    def write_tech(
        self,
        tech_dir: PathType,
        lyp_filename: str = "layers.lyp",
        lyt_filename: str = "tech.lyt",
        d25_filename: Optional[str] = None,
        mebes_config: Optional[dict] = None,
    ) -> None:
        """Write technology files into 'tech_dir'.

        Args:
            tech_dir: Where to write the technology files to.
            lyp_filename: Name of the layer properties file.
            lyt_filename: Name of the layer technology file.
            d25_filename: Name of the 2.5D stack file (only works on KLayout >= 0.28). Defaults to self.name.
            mebes_config: A dictionary specifying the KLayout mebes reader config.

        """
        from gdsfactory.utils.xml_utils import make_pretty_xml

        d25_filename = d25_filename or f"{self.name}.lyd25"

        tech_path = pathlib.Path(tech_dir)
        lyp_path = tech_path / lyp_filename
        lyt_path = tech_path / lyt_filename
        d25_path = tech_path / "d25" / d25_filename
        d25_dir = tech_path / "d25"
        d25_dir.mkdir(exist_ok=True, parents=True)

        if not self.technology.name:
            self.technology.name = self.name

        self.technology.layer_properties_file = lyp_path.name

        if self.layer_views:
            self.layer_views.to_lyp(lyp_path)
            print(f"Wrote {str(lyp_path)!r}")

        root = ET.XML(self.technology.to_xml().encode("utf-8"))

        # KLayout tech doesn't include mebes config, so add it after lefdef config:
        if not mebes_config:
            mebes_config = {
                "invert": False,
                "subresolution": True,
                "produce-boundary": True,
                "num-stripes-per-cell": 64,
                "num-shapes-per-cell": 0,
                "data-layer": 1,
                "data-datatype": 0,
                "data-name": "DATA",
                "boundary-layer": 0,
                "boundary-datatype": 0,
                "boundary-name": "BORDER",
                "layer-map": "layer_map()",
                "create-other-layers": True,
            }
        mebes = ET.Element("mebes")
        for k, v in mebes_config.items():
            v = str(v).lower() if isinstance(v, bool) else str(v)
            ET.SubElement(mebes, k).text = v

        reader_opts = root.find("reader-options")
        lefdef_idx = list(reader_opts).index(reader_opts.find("lefdef"))
        reader_opts.insert(lefdef_idx + 1, mebes)

        if self.layer_stack:
            dbu = len(str(self.technology.dbu).split(".")[-1])
            d25_script = (
                prefix_d25
                + self.layer_stack.get_klayout_3d_script(
                    layer_views=self.layer_views,
                    dbu=dbu,
                )
                + suffix_d25
            )
            d25_path.write_bytes(d25_script.encode("utf-8"))
            print(f"Wrote {str(d25_path)!r}")

        self._define_connections(root)

        lyt_path.write_bytes(make_pretty_xml(root))
        print(f"Wrote {str(lyt_path)!r}")

    def _define_connections(self, root) -> None:
        if not self.connectivity:
            return
        src_element = [e for e in list(root) if e.tag == "connectivity"]
        if len(src_element) != 1:
            raise KeyError("Could not get a single index for the src element.")
        src_element = src_element[0]
        layers = set()
        for layer_name_c1, layer_name_via, layer_name_c2 in self.connectivity:
            connection = ",".join([layer_name_c1, layer_name_via, layer_name_c2])

            layers.add(layer_name_c1)
            layers.add(layer_name_via)
            layers.add(layer_name_c2)

            ET.SubElement(src_element, "connection").text = connection

        if self.layer_map:
            for layer in layers:
                ET.SubElement(
                    src_element, "symbols"
                ).text = (
                    f"{layer}='{self.layer_map[layer][0]}/{self.layer_map[layer][1]}'"
                )

    class Config:
        """Allow db.Technology type."""

        arbitrary_types_allowed = True


layer_views = LayerViews.from_lyp(str(PATH.klayout_lyp))


def yaml_test() -> None:
    tech_dir = PATH.repo / "extra" / "test_tech"

    # Load from existing layer properties file
    lyp = LayerViews.from_lyp(str(PATH.klayout_lyp))
    print("Loaded from .lyp", lyp)

    # Export layer properties to yaml files
    layer_yaml = str(tech_dir / "layers.yml")
    lyp.to_yaml(layer_yaml)

    # Load layer properties from yaml files and check that they're the same
    lyp_loaded = LayerViews.from_yaml(layer_yaml)
    print("Loaded from .yaml", lyp_loaded)

    assert lyp_loaded == lyp


if __name__ == "__main__":
    from gdsfactory.generic_tech import LAYER, LAYER_STACK

    lyp = LayerViews(PATH.klayout_yaml)
    # lyp = LayerViews.from_lyp(str(PATH.klayout_yaml))

    # str_xml = open(PATH.klayout_tech / "tech.lyt").read()
    # new_tech = db.Technology.technology_from_xml(str_xml)
    # generic_tech = KLayoutTechnology(layer_views=lyp)
    connectivity = [
        ("NPP", "VIAC", "M1"),
        ("PPP", "VIAC", "M1"),
        ("M1", "VIA1", "M2"),
        ("M2", "VIA2", "M3"),
    ]

    c = generic_tech = KLayoutTechnology(
        name="generic_tech", layer_views=lyp, connectivity=connectivity, layer_map=LAYER
    )
    tech_dir = PATH.klayout_tech
    # tech_dir = pathlib.Path("/home/jmatres/.klayout/salt/gdsfactory/tech/")
    tech_dir.mkdir(exist_ok=True, parents=True)
    generic_tech.write_tech(tech_dir=tech_dir, layer_stack=LAYER_STACK)

    # yaml_test()



from __future__ import annotations

import numpy as np

import gdsfactory.simulation.gtidy3d as gt


def test_neff() -> None:
    wg = gt.modes.Waveguide(
        wavelength=1.55,
        core_width=0.5,
        core_thickness=0.22,
        core_material="si",
        clad_material="sio2",
        cache=False,
    )
    n_eff = wg.n_eff[0].real
    assert np.isclose(n_eff, 2.447, rtol=0.1), n_eff


def test_neff_high_accuracy() -> None:
    wg = gt.modes.Waveguide(
        wavelength=1.55,
        core_width=0.5,
        core_thickness=0.22,
        core_material="si",
        clad_material="sio2",
        grid_resolution=40,
        cache=False,
    )
    n_eff = wg.n_eff[0].real
    assert np.isclose(n_eff, 2.447, rtol=0.01), n_eff



"""Wires for electrical manhattan routes."""

from __future__ import annotations

from functools import partial

import numpy as np

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.straight import straight
from gdsfactory.typings import CrossSectionSpec

wire_straight = partial(straight, cross_section="metal_routing")


@gf.cell
def wire_corner(
    cross_section: CrossSectionSpec = "metal_routing", **kwargs
) -> Component:
    """Returns 45 degrees electrical corner wire.

    Args:
        cross_section: spec.
        kwargs: cross_section settings.
    """
    x = gf.get_cross_section(cross_section, **kwargs)
    layer = x.layer
    width = x.width

    c = Component()
    a = width / 2
    xpts = [-a, a, a, -a]
    ypts = [-a, -a, a, a]

    c.add_polygon([xpts, ypts], layer=layer)
    c.add_port(
        name="e1",
        center=(-a, 0),
        width=width,
        orientation=180,
        layer=layer,
        port_type="electrical",
    )
    c.add_port(
        name="e2",
        center=(0, a),
        width=width,
        orientation=90,
        layer=layer,
        port_type="electrical",
    )
    c.info["length"] = width
    c.info["dy"] = width
    return c


@gf.cell
def wire_corner45(
    cross_section: CrossSectionSpec = "metal_routing", radius: float = 10, **kwargs
) -> Component:
    """Returns 90 degrees electrical corner wire.

    Args:
        cross_section: spec.
        kwargs: cross_section settings.
    """
    x = gf.get_cross_section(cross_section, **kwargs)
    layer = x.layer
    width = x.width
    radius = x.radius if radius is None else radius
    if radius is None:
        raise ValueError(
            "Radius needs to be specified in wire_corner45 or in the cross_section."
        )

    c = Component()

    a = width / 2

    xpts = [0, radius + a, radius + a, -np.sqrt(2) * width]
    ypts = [-a, radius, radius + np.sqrt(2) * width, -a]

    c.add_polygon([xpts, ypts], layer=layer)
    c.add_port(
        name="e1",
        center=(0, 0),
        width=width,
        orientation=180,
        layer=layer,
        port_type="electrical",
    )
    c.add_port(
        name="e2",
        center=(radius, radius),
        width=width,
        orientation=90,
        layer=layer,
        port_type="electrical",
    )
    c.info["length"] = np.sqrt(2) * radius
    return c


@gf.cell
def wire_corner_sections(
    cross_section: CrossSectionSpec = "metal_routing",
    **kwargs,
) -> Component:
    """Returns 90 degrees electrical corner wire, where all cross_section sections properly represented.

    Works well with symmetric cross_sections, not quite ready for asymmetric.

    Args:
        cross_section: spec.
        kwargs: cross_section settings.
    """
    x = gf.get_cross_section(cross_section, **kwargs)

    xmin, ymax = x.get_xmin_xmax()

    main_section = gf.cross_section.Section(
        width=x.width,
        layer=x.layer,
        offset=x.offset,
    )

    all_sections = [main_section]
    all_sections.extend(x.sections)

    c = Component()

    for section in all_sections:
        layer = section.layer
        width = section.width
        offset = section.offset
        b = width / 2

        xpts = [xmin, offset - b, offset - b, offset + b, offset + b, xmin]
        ypts = [
            -offset + b,
            -offset + b,
            ymax,
            ymax,
            -offset - b,
            -offset - b,
        ]

        c.add_polygon([xpts, ypts], layer=layer)

    c.add_port(
        name="e1",
        center=(xmin, -(xmin + ymax) / 2),
        orientation=180,
        cross_section=x,
        layer=x.layer,
    )
    c.add_port(
        name="e2",
        center=((xmin + ymax) / 2, ymax),
        orientation=90,
        cross_section=x,
        layer=x.layer,
    )
    c.info["length"] = ymax - xmin
    c.info["dy"] = ymax - xmin
    return c


if __name__ == "__main__":
    # c = wire_straight()
    # c = wire_corner()
    section1 = gf.cross_section.Section(width=0.4, layer="M1", offset=0.5)
    section2 = gf.cross_section.Section(width=0.4, layer="M1", offset=-0.5)
    xsection = gf.cross_section.cross_section(
        width=0.4, offset=0, layer="M1", sections=[section1, section2]
    )
    from gdsfactory.cross_section import metal_slotted

    c = wire_corner_sections(cross_section=metal_slotted)
    c.show(show_ports=True)
    # c.pprint_ports()
    # c.pprint()

    # port1 = gf.Port(name="init", orientation=270, center=(0, 0), cross_section=metal_slotted)
    # port2 = gf.Port(
    #     name="final", orientation=0, center=(-20, -20), cross_section=metal_slotted
    # )

    # c = gf.Component()
    # route = gf.routing.get_route_electrical(
    #     input_port=port1,
    #     output_port=port2,
    #     bend=wire_corner_sections(cross_section=xsection),
    #     cross_section=xsection,
    # )
    # c.add(route.references)
    # c.show()

    # print(yaml.dump(c.to_dict()))



import base64
import importlib
import os
from pathlib import Path
from typing import Optional

import orjson
from fastapi import FastAPI, Form, Request, status
from gdsfactory.plugins.web.middleware import ProxiedHeadersMiddleware
from fastapi.responses import HTMLResponse, RedirectResponse
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates
from loguru import logger
from starlette.routing import WebSocketRoute

import gdsfactory as gf
from gdsfactory.config import PATH
from gdsfactory.plugins.web.server import LayoutViewServerEndpoint, get_layout_view

module_path = Path(__file__).parent.absolute()

app = FastAPI(
    routes=[WebSocketRoute("/view/{cell_name}/ws", endpoint=LayoutViewServerEndpoint)]
)
app.add_middleware(ProxiedHeadersMiddleware)
# app = FastAPI()
app.mount("/static", StaticFiles(directory=PATH.web / "static"), name="static")

# gdsfiles = StaticFiles(directory=home_path)
# app.mount("/gds_files", gdsfiles, name="gds_files")
templates = Jinja2Templates(directory=PATH.web / "templates")


def load_pdk() -> gf.Pdk:
    pdk = os.environ.get("PDK", "generic")

    if pdk == "generic":
        active_pdk = gf.get_active_pdk()
    else:
        active_module = importlib.import_module(pdk)
        active_pdk = active_module.PDK
        active_pdk.activate()
    return active_pdk


def get_url(request: Request) -> str:
    port_mod = ""
    if request.url.port is not None and len(str(request.url).split(".")) < 3:
        port_mod = ":" + str(request.url.port)

    hostname = request.url.hostname

    if "github" in hostname:
        port_mod = ""

    url = str(
        request.url.scheme
        + "://"
        + (hostname or "localhost")
        + port_mod
        + request.url.path
    )

    return url


@app.get("/", response_class=HTMLResponse)
async def root(request: Request):
    if "preview.app.github" in str(request.url):
        return RedirectResponse(str(request.url).replace(".preview", ""))
    active_pdk = load_pdk()
    pdk_name = active_pdk.name
    components = list(active_pdk.cells.keys())
    return templates.TemplateResponse(
        "index.html",
        {
            "request": request,
            "title": "Main",
            "pdk_name": pdk_name,
            "components": sorted(components),
        },
    )


LOADED_COMPONENTS = {}


@app.get("/view/{cell_name}", response_class=HTMLResponse)
async def view_cell(request: Request, cell_name: str, variant: Optional[str] = None):
    if "preview.app.github" in str(request.url):
        return RedirectResponse(str(request.url).replace(".preview", ""))

    if variant in LOADED_COMPONENTS:
        component = LOADED_COMPONENTS[variant]
    else:
        component = gf.get_component(cell_name)
    layout_view = get_layout_view(component)
    pixel_data = layout_view.get_pixels_with_options(800, 400).to_png_data()
    # pixel_data = layout_view.get_screenshot_pixels().to_png_data()
    b64_data = base64.b64encode(pixel_data).decode("utf-8")
    return templates.TemplateResponse(
        "viewer.html",
        {
            "request": request,
            "cell_name": str(cell_name),
            "variant": variant,
            "title": "Viewer",
            "initial_view": b64_data,
            "component": component,
            "url": get_url(request),
        },
    )


def _parse_value(value: str):
    if value.startswith("{") or value.startswith("["):
        try:
            return orjson.loads(value.replace("'", '"'))
        except orjson.JSONDecodeError as e:
            raise ValueError(f"Unable to decode parameter value, {value}: {e.msg}")
    else:
        return value


@app.post("/update/{cell_name}")
async def update_cell(request: Request, cell_name: str):
    data = await request.form()
    changed_settings = {k: _parse_value(v) for k, v in data.items() if v != ""}
    new_component = gf.get_component(
        {"component": cell_name, "settings": changed_settings}
    )
    LOADED_COMPONENTS[new_component.name] = new_component
    logger.info(data)
    return RedirectResponse(
        f"/view/{cell_name}?variant={new_component.name}",
        status_code=status.HTTP_302_FOUND,
    )


@app.post("/search", response_class=RedirectResponse)
async def search(name: str = Form(...)):
    logger.info(f"Searching for {name}...")
    try:
        gf.get_component(name)
    except ValueError:
        return RedirectResponse("/", status_code=status.HTTP_404_NOT_FOUND)
    logger.info(f"Successfully found {name}! Redirecting...")
    return RedirectResponse(f"/view/{name}", status_code=status.HTTP_302_FOUND)


if __name__ == "__main__":
    pdk = load_pdk()



from typing import List, Tuple

from starlette.types import ASGIApp, Receive, Scope, Send

Headers = List[Tuple[bytes, bytes]]

# original developed by researchers at university of cambridge:
# https://pypi.org/project/fastapi-proxiedheadersmiddleware/

# modification to resolve Connection: keep-alive vs. upgrade conflict
# conflict described here: https://github.com/orgs/community/discussions/57596
# doesn't seem to work as intended, as the upgrade decision is made
# before the request goes through the middleware


class ProxiedHeadersMiddleware:
    """
    A middleware that modifies the request to ensure that FastAPI uses the
    X-Forwarded-* headers when creating URLs used to reference this application.

    We are very permissive in allowing all X-Forwarded-* headers to be used, as
    we know that this API will be published behind the API Gateway, and is
    therefore not prone to redirect hijacking.

    """

    def __init__(self, app: ASGIApp):
        self.app = app

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        scope["headers"] = self.remap_headers(scope.get("headers", {}))

        await self.app(scope, receive, send)
        return

    def remap_headers(self, source: Headers) -> Headers:
        """
        Map X-Forwarded-Host to host and X-Forwarded-Prefix to prefix.

        """
        upgrade = len(
            [q for p, q in source if "connection" in str(p) and "upgrade" in str(q)]
        )

        source = dict(source)
        if upgrade:  # resolve conflict with priority of upgrade
            source[b"connection"] = b"upgrade"

        if b"x-forwarded-host" in source:
            source.update({b"host": source[b"x-forwarded-host"]})
            source.pop(b"x-forwarded-host")

        if b"x-forwarded-prefix" in source:
            source.update({b"host": source[b"host"] + source[b"x-forwarded-prefix"]})
            source.pop(b"x-forwarded-prefix")

        source = [(k, v) for k, v in source.items()]

        return source



import time
from typing import Tuple

import numpy as np
from femwell.maxwell.waveguide import compute_modes, Modes
from skfem import (
    Basis,
    ElementTriN2,
    ElementTriP0,
    ElementTriP2,
    Mesh,
)

import gdsfactory as gf
from gdsfactory.pdk import get_layer_stack, get_material_index
from gdsfactory.technology import LayerStack
from gdsfactory.typings import ComponentSpec, CrossSectionSpec, PathType

mesh_filename = "mesh.msh"


def load_mesh_basis(mesh_filename: PathType):
    mesh = Mesh.load(mesh_filename)
    basis = Basis(mesh, ElementTriN2() * ElementTriP2())
    basis0 = basis.with_element(ElementTriP0())
    return mesh, basis0


def compute_cross_section_modes(
    cross_section: CrossSectionSpec,
    layerstack: LayerStack,
    wavelength: float = 1.55,
    num_modes: int = 4,
    order: int = 1,
    radius: float = np.inf,
    wafer_padding: float = 2.0,
    **kwargs,
) -> Modes:
    """Calculate effective index of a cross-section.

    Defines a "straight" component of the cross_section, and calls compute_component_slice_modes.

    Args:
        cross_section:
        layerstack:
        wavelength: in um.
        num_modes: to compute.
        order: order of the mesh elements. 1: linear, 2: quadratic.
        radius: defaults to inf.
        wafer_padding: in um.

    Keyword Args:
        solver: can be slepc or scipy.
        resolutions (Dict): Pairs {"layername": {"resolution": float, "distance": "float}}
            to roughly control mesh refinement within and away from entity, respectively.
        mesh_scaling_factor (float): factor multiply mesh geometry by.
        default_resolution_min (float): gmsh minimal edge length.
        default_resolution_max (float): gmsh maximal edge length.
        background_tag (str): name of the background layer to add (default: no background added).
        background_padding (Tuple): [xleft, ydown, xright, yup] distances to add to the components and to fill with background_tag.
        global_meshsize_array: np array [x,y,z,lc] to parametrize the mesh.
        global_meshsize_interpolant_func: interpolating function for global_meshsize_array.
        extra_shapes_dict: Optional[OrderedDict] of {key: geo} with key a label and geo a shapely (Multi)Polygon or (Multi)LineString of extra shapes to override component.
        merge_by_material: boolean, if True will merge polygons from layers with the same layer.material. Physical keys will be material in this case.

    """
    # Get meshable component from cross-section
    c = gf.components.straight(length=10, cross_section=cross_section)
    bounds = c.bbox
    dx = np.diff(bounds[:, 0])[0]

    xsection_bounds = [
        [dx / 2, bounds[0, 1] - wafer_padding],
        [dx / 2, bounds[1, 1] + wafer_padding],
    ]

    # Mesh as component
    return compute_component_slice_modes(
        component=c,
        xsection_bounds=xsection_bounds,
        layerstack=layerstack,
        wavelength=wavelength,
        num_modes=num_modes,
        order=order,
        radius=radius,
        wafer_padding=wafer_padding,
        **kwargs,
    )


def compute_component_slice_modes(
    component: ComponentSpec,
    xsection_bounds: Tuple[Tuple[float, float], Tuple[float, float]],
    layerstack: LayerStack,
    wavelength: float = 1.55,
    num_modes: int = 4,
    order: int = 1,
    radius: float = np.inf,
    wafer_padding: float = 2.0,
    solver: str = "scipy",
    **kwargs,
) -> Modes:
    """Calculate effective index of component slice.

    Args:
        component: gdsfactory component.
        xsection_bounds: xy line defining where to take component cross_section.
        layerstack: gdsfactory layerstack.
        wavelength: wavelength (um).
        num_modes: number of modes to return.
        order: order of the mesh elements. 1: linear, 2: quadratic.
        radius: bend radius of the cross-section.
        wafer_padding: padding beyond bbox to add to WAFER layers.
        bend_radius: bend radius for mode solving (typically called from cross_section)
        solver: can be slepc or scipy.

    Keyword Args:
        resolutions (Dict): Pairs {"layername": {"resolution": float, "distance": "float}}
            to roughly control mesh refinement within and away from entity, respectively.
        mesh_scaling_factor (float): factor multiply mesh geometry by.
        default_resolution_min (float): gmsh minimal edge length.
        default_resolution_max (float): gmsh maximal edge length.
        background_tag (str): name of the background layer to add (default: no background added).
        background_padding (Tuple): [xleft, ydown, xright, yup] distances to add to the components and to fill with background_tag.
        global_meshsize_array: np array [x,y,z,lc] to parametrize the mesh.
        global_meshsize_interpolant_func: interpolating function for global_meshsize_array.
        extra_shapes_dict: Optional[OrderedDict] of {key: geo} with key a label and geo a shapely (Multi)Polygon or (Multi)LineString of extra shapes to override component.
        merge_by_material: boolean, if True will merge polygons from layers with the same layer.material. Physical keys will be material in this case.
    """

    # Mesh
    mesh = component.to_gmsh(
        type="uz",
        xsection_bounds=xsection_bounds,
        layer_stack=layerstack,
        filename=mesh_filename,
        wafer_padding=wafer_padding,
        **kwargs,
    )

    # Assign materials to mesh elements
    mesh, basis0 = load_mesh_basis(mesh_filename)
    epsilon = basis0.zeros(dtype=complex)
    for layername, layer in layerstack.layers.items():
        if layername in mesh.subdomains.keys():
            epsilon[basis0.get_dofs(elements=layername)] = (
                get_material_index(layer.material, wavelength) ** 2
            )
        if "background_tag" in kwargs:
            epsilon[basis0.get_dofs(elements=kwargs["background_tag"])] = (
                get_material_index(kwargs["background_tag"], wavelength) ** 2
            )

    # Mode solve
    modes = compute_modes(
        basis0,
        epsilon,
        wavelength=wavelength,
        mu_r=1,
        num_modes=num_modes,
        order=order,
        radius=radius,
        solver=solver,
    )
    return modes


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    PDK = gf.get_generic_pdk()
    PDK.activate()

    start = time.time()

    filtered_layerstack = LayerStack(
        layers={
            k: get_layer_stack().layers[k]
            for k in (
                "core",
                "clad",
                "slab90",
                "box",
            )
        }
    )

    filtered_layerstack.layers["core"].thickness = 0.2

    resolutions = {
        "core": {"resolution": 0.02, "distance": 2},
        "clad": {"resolution": 0.2, "distance": 1},
        "box": {"resolution": 0.2, "distance": 1},
        "slab90": {"resolution": 0.05, "distance": 1},
    }

    cross_section = False
    cross_section = True

    if cross_section:
        modes = compute_cross_section_modes(
            cross_section="rib",
            layerstack=filtered_layerstack,
            wavelength=1.55,
            num_modes=4,
            order=1,
            radius=np.inf,
            resolutions=resolutions,
        )
        mode = modes[0]
        mode.show(mode.E.real, colorbar=True, direction="x")
    else:
        component = gf.components.coupler_full(dw=0)
        component.show()

        modes = compute_component_slice_modes(
            component,
            [[0, -3], [0, 3]],
            layerstack=filtered_layerstack,
            wavelength=1.55,
            num_modes=4,
            order=1,
            radius=np.inf,
            mesh_filename="./mesh.msh",
            resolutions=resolutions,
        )

        print(modes)
        print(modes[0].te_fraction)

        modes[0].show(np.real(modes[0].E))
        modes[0].show(np.imag(modes[0].E))

        modes[0].show(np.real(modes[0].H))
        modes[0].show(np.imag(modes[0].H))

        integrals = np.zeros((len(modes),) * 2, dtype=complex)

        for i in range(len(modes)):
            for j in range(len(modes)):
                integrals[i, j] = modes[i].calculate_overlap(modes[j])

        plt.imshow(np.real(integrals))
        plt.colorbar()
        plt.show()

        # Create basis to select a certain simulation extent
        def sel_fun(x):
            return (x[0] < 0) * (x[0] > -1) * (x[1] > 0) * (x[1] < 0.5)

        print(modes.sorted(lambda mode: mode.calculate_power(elements=sel_fun)))



from __future__ import annotations

from functools import partial
from typing import Dict, Union

import tidy3d as td
from tidy3d.components.medium import PoleResidue
from tidy3d.components.types import ComplexNumber

MATERIAL_NAME_TO_MEDIUM = {
    "si": td.material_library["cSi"]["Li1993_293K"],
    "sio2": td.material_library["SiO2"]["Horiba"],
    "sin": td.material_library["Si3N4"]["Luke2015PMLStable"],
}


def get_epsilon(
    name_or_index: Union[str, float],
    wavelength: float = 1.55,
    material_name_to_medium: Dict[str, PoleResidue] = MATERIAL_NAME_TO_MEDIUM,
) -> ComplexNumber:
    """Return permittivity from material database.

    Args:
        name_or_index: material name or refractive index.
        wavelength: wavelength (um).
        material_name_to_medium: map name to medium.

    """
    medium = get_medium(
        name_or_index=name_or_index, material_name_to_medium=material_name_to_medium
    )
    frequency = td.C_0 / wavelength
    return medium.eps_model(frequency)


def get_index(
    name_or_index: Union[str, float],
    wavelength: float = 1.55,
    material_name_to_medium: Dict[str, PoleResidue] = MATERIAL_NAME_TO_MEDIUM,
) -> float:
    """Return refractive index from material database.

    Args:
        wavelength: wavelength (um).
        name_or_index: material name or refractive index.
        material_name_to_medium: map name to medium.

    """
    eps_complex = get_epsilon(
        wavelength=wavelength,
        name_or_index=name_or_index,
        material_name_to_medium=material_name_to_medium,
    )
    n, _ = td.Medium.eps_complex_to_nk(eps_complex)
    return n


def get_nk(
    name_or_index: Union[str, float],
    wavelength: float = 1.55,
    material_name_to_medium: Dict[str, PoleResidue] = MATERIAL_NAME_TO_MEDIUM,
) -> float:
    """Return refractive index and optical extinction coefficient
    from material database.

    Args:
        wavelength: wavelength (um).
        name_or_index: material name or refractive index.
        material_name_to_medium: map name to medium.

    """
    eps_complex = get_epsilon(
        wavelength=wavelength,
        name_or_index=name_or_index,
        material_name_to_medium=material_name_to_medium,
    )
    n, k = td.Medium.eps_complex_to_nk(eps_complex)
    return n, k


def get_medium(
    name_or_index: Union[str, float],
    material_name_to_medium: Dict[str, PoleResidue] = MATERIAL_NAME_TO_MEDIUM,
) -> td.Medium:
    """Return Medium from materials database.

    Args:
        name_or_index: material name or refractive index.
        material_name_to_medium: map name to medium.

    """
    name_or_index = (
        name_or_index.lower() if isinstance(name_or_index, str) else name_or_index
    )

    if isinstance(name_or_index, (int, float)):
        m = td.Medium(permittivity=name_or_index**2)
    elif name_or_index in material_name_to_medium:
        m = material_name_to_medium[name_or_index]
    else:
        materials = list(material_name_to_medium.keys())

        raise ValueError(f"Material {name_or_index!r} not in {materials}")

    return m


si = partial(get_index, "si")
sio2 = partial(get_index, "sio2")
sin = partial(get_index, "sin")

materials = dict(si=si, sio2=sio2, sin=sin)


if __name__ == "__main__":
    print(si(1.55))
    print(si(1.31))
    # print(get_index(name_or_index="cSi"))
    # print(get_index(name_or_index=3.4))
    # m = get_medium(name_or_index="SiO2")
    # m = td.Medium(permittivity=1.45 ** 2)



import numpy as np

import gdsfactory as gf
from gdsfactory.generic_tech import LAYER_STACK
from gdsfactory.simulation.fem.mode_solver import compute_cross_section_modes, Modes
from gdsfactory.technology import LayerStack

NUM_MODES = 1

PDK = gf.get_generic_pdk()
PDK.activate()


def compute_modes(
    overwrite: bool = True, with_cache: bool = False, num_modes: int = NUM_MODES
) -> Modes:
    filtered_layerstack = LayerStack(
        layers={
            k: LAYER_STACK.layers[k]
            for k in (
                "core",
                "clad",
                "slab90",
                "box",
            )
        }
    )

    filtered_layerstack.layers["core"].thickness = 0.2

    resolutions = {
        "core": {"resolution": 0.02, "distance": 2},
        "clad": {"resolution": 0.2, "distance": 1},
        "box": {"resolution": 0.2, "distance": 1},
        "slab90": {"resolution": 0.05, "distance": 1},
    }
    return compute_cross_section_modes(
        cross_section="rib",
        layerstack=filtered_layerstack,
        wavelength=1.55,
        num_modes=num_modes,
        order=1,
        radius=np.inf,
        resolutions=resolutions,
        overwrite=overwrite,
        with_cache=with_cache,
    )


def test_compute_cross_section_mode() -> None:
    modes = compute_modes()
    assert len(modes) == NUM_MODES, len(modes)


if __name__ == "__main__":
    test_compute_cross_section_mode()



"""based on phidl.geometry."""
from __future__ import annotations

from functools import partial
from typing import Optional

import numpy as np

import gdsfactory as gf
from gdsfactory.components.rectangle import rectangle
from gdsfactory.components.text import text
from gdsfactory.typings import Anchor, ComponentSpec, LayerSpec

big_square = partial(rectangle, size=(1300, 2600))


@gf.cell
def die_bbox(
    component: ComponentSpec = big_square,
    street_width: float = 100.0,
    street_length: Optional[float] = None,
    die_name: Optional[str] = None,
    text_size: float = 100.0,
    text_anchor: Anchor = "sw",
    layer: LayerSpec = "M3",
    padding: float = 10.0,
) -> gf.Component:
    """Returns component with boundary box frame around it.

    Perfect for defining the boundary of the chip/die
    it can also add a label with the name of the die.
    similar to die and bbox.

    Args:
        component: to frame.
        street_width: Width of the boundary box.
        street_length: length of the boundary box.
        die_name: Label text.
        text_size: Label text size.
        text_anchor: {'nw', 'nc', 'ne', 'sw', 'sc', 'se'} text location.
        layer: Specific layer(s) to put polygon geometry on.
        padding: adds padding.
    """
    D = gf.Component()
    component = gf.get_component(component)

    D.copy_child_info(component)
    cref = D.add_ref(component)
    cref.x = 0
    cref.y = 0
    size = cref.size
    sx, sy = size[0] / 2, size[1] / 2

    sx += street_width + padding
    sy += street_width + padding

    street_length = street_length or max([sx, sy])

    xpts = np.array(
        [
            sx,
            sx,
            sx - street_width,
            sx - street_width,
            sx - street_length,
            sx - street_length,
        ]
    )
    ypts = np.array(
        [
            sy,
            sy - street_length,
            sy - street_length,
            sy - street_width,
            sy - street_width,
            sy,
        ]
    )
    D.add_polygon([+xpts, +ypts], layer=layer)
    D.add_polygon([-xpts, +ypts], layer=layer)
    D.add_polygon([+xpts, -ypts], layer=layer)
    D.add_polygon([-xpts, -ypts], layer=layer)

    if die_name:
        t = D.add_ref(text(text=die_name, size=text_size, layer=layer))

        d = street_width + 20
        if text_anchor == "nw":
            t.xmin, t.ymax = [-sx + d, sy - d]
        elif text_anchor == "nc":
            t.x, t.ymax = [0, sy - d]
        elif text_anchor == "ne":
            t.xmax, t.ymax = [sx - d, sy - d]
        if text_anchor == "sw":
            t.xmin, t.ymin = [-sx + d, -sy + d]
        elif text_anchor == "sc":
            t.x, t.ymin = [0, -sy + d]
        elif text_anchor == "se":
            t.xmax, t.ymin = [sx - d, -sy + d]

    return D


if __name__ == "__main__":
    mask = gf.components.array(rows=10, columns=10)
    # c = die_bbox(component=mask, die_name="chip99")
    c = die_bbox()
    # c.show(show_ports=True)



from __future__ import annotations

import numpy as np

import gdsfactory as gf
from gdsfactory.typings import Component
from gdsfactory.samples.big_device import big_device

debug = False
nlabels = 12


def mzi_te(**kwargs) -> Component:
    """Returns MZI with TE grating couplers."""
    gc = gf.c.grating_coupler_elliptical_tm()
    c = gf.c.mzi_phase_shifter_top_heater_metal(delta_length=40)
    c = gf.routing.add_fiber_array(c, grating_coupler=gc, **kwargs)
    c = gf.routing.add_electrical_pads_shortest(c)
    return c


def spirals() -> Component:
    cm = 10e3
    lengths = np.array([2, 4, 6]) * cm
    spirals = [gf.components.spiral_inner_io(length=length) for length in lengths]
    spirals_te = [gf.components.add_grating_couplers(c) for c in spirals]
    return gf.pack(spirals_te)[0]


@gf.cell
def demo_pack() -> Component:
    """Sample reticle."""
    c = gf.Component()
    c << gf.c.rectangle(size=(22e3, 22e3), layer=gf.LAYER.FLOORPLAN)
    components = [mzi_te()] + [spirals()] + [gf.routing.add_fiber_array(big_device())]
    c << gf.pack(components)[0]
    return c


if __name__ == "__main__":
    # c = spirals()
    c = demo_pack()
    c.show()



from typing import Dict

from dagster import ConfigurableResource, Definitions, asset


class ReaderResource(ConfigurableResource):
    value: str


@asset
def design(reader: ReaderResource) -> Dict:
    # read_based_on_config()
    return {"design:": reader.value}


@asset
def verification():
    return {"verification:": 1}


@asset
def manufacturing():
    return {"manufacturing:": 1}


@asset
def validation():
    return {"validation:": 1}


@asset
def structure_layer(
    context,
    design,
    verification,
    manufacturing,
    validation,
):
    combined = {
        **design,
        **verification,
        **manufacturing,
        **validation,
    }
    context.log.info(f"Combined data {combined}")
    return combined


defs = Definitions(
    assets=[
        design,
        verification,
        manufacturing,
        validation,
        structure_layer,
    ],
    resources={"reader": ReaderResource(value="configured-value")},
)
